"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable functional/prefer-immutable-types */
const utils_1 = require("@typescript-eslint/utils");
const common_1 = require("./common");
const is_immutable_type_1 = require("is-immutable-type");
/**
 * An ESLint rule to ban decreasing the immutability level, as determined by https://github.com/RebeccaStevens/is-immutable-type#definitions.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noDecreasingImmutability = (0, common_1.createRule)({
    name: "no-decreasing-immutability",
    meta: {
        type: "problem",
        docs: {
            description: "Forbids decreasing the immutability level.",
            recommended: "error",
        },
        messages: {
            errorStringGeneric: 'Do not decrease the immutability level of a value from "{{ sourceImmutability }}"" to "{{ destinationImmutability }}".',
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        const decreasingImmutabilityViolation = (checker, destination, source) => {
            const destinationImmutability = (0, is_immutable_type_1.getTypeImmutability)(checker, destination);
            const sourceImmutability = (0, is_immutable_type_1.getTypeImmutability)(checker, source);
            // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
            return (0, is_immutable_type_1.isUnknown)(destinationImmutability) ||
                (0, is_immutable_type_1.isUnknown)(sourceImmutability) ||
                destinationImmutability >= sourceImmutability
                ? undefined
                : {
                    destinationImmutability: is_immutable_type_1.Immutability[destinationImmutability],
                    sourceImmutability: is_immutable_type_1.Immutability[sourceImmutability],
                };
        };
        const reportViolation = (checker, destination, source, node
        // eslint-disable-next-line functional/no-return-void
        ) => {
            const maybeViolation = decreasingImmutabilityViolation(checker, destination, source);
            // eslint-disable-next-line functional/no-conditional-statements
            if (maybeViolation !== undefined) {
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                    data: maybeViolation,
                });
            }
        };
        return {
            // eslint-disable-next-line functional/no-return-void
            AssignmentExpression: (node) => {
                const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.left);
                const rightTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.right);
                const leftType = checker.getTypeAtLocation(leftTsNode);
                const rightType = checker.getTypeAtLocation(rightTsNode);
                // eslint-disable-next-line functional/no-expression-statements
                reportViolation(checker, leftType, rightType, node);
            },
        };
    },
    defaultOptions: [],
});
exports.default = noDecreasingImmutability;
